/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.taxonomy.spi;

import it.tidalwave.bluebill.mobile.taxonomy.text.FastNameMatcher;
import it.tidalwave.bluebill.mobile.taxonomy.text.ExactSubstringNameMatcher;
import it.tidalwave.bluebill.mobile.taxonomy.text.CamelCaseMatcher;
import it.tidalwave.bluebill.mobile.preferences.NameMatcher;
import it.tidalwave.observation.bluebill.ObservationClipboard;
import it.tidalwave.bluebill.mobile.preferences.TaxonomyPreferences;
import it.tidalwave.bluebill.mobile.taxonomy.TaxonHistory;
import it.tidalwave.observation.ObservationManager;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.text.Collator;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import it.tidalwave.bluebill.taxonomy.mobile.Taxonomy;
import it.tidalwave.bluebill.taxonomy.mobile.TaxonomyFactory;
import it.tidalwave.bluebill.taxonomy.mobile.impl.SimpleTaxonomyFactory;
import it.tidalwave.mobile.preferences.PreferencesAdapter;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class TaxonomyPreferencesSupport implements TaxonomyPreferences
  {
    private static final String CLASS = TaxonomyPreferencesSupport.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    public static final String PREF_RENDER_SCIENTIFIC_NAMES = "renderScientificNames";
    public static final String PREF_PRIMARY_LOCALE = "taxonomyPrimaryLocale";
    public static final String PREF_SECONDARY_LOCALE = "taxonomySecondaryLocale";
    public static final String PREF_NAME_MATCHER = "nameMatcher/3";
    public static final String PREF_TAXONOMY = "taxonomySchema";
   
    private static final String DEFAULT_NAME_MATCHER = "FastNameMatcherWithInitial";

    @CheckForNull
    protected Taxonomy taxonomy;

    protected final List<Locale> taxonLocales = new ArrayList<Locale>();

    protected Locale primaryTaxonLocale = Locale.ENGLISH;

    protected NameMatcher nameMatcher;

    protected Collator taxonCollator = Collator.getInstance();

    protected final Map<String, NameMatcher> matcherMapByKey = new HashMap<String, NameMatcher>();

    protected final PreferencesAdapter preferencesAdapter = Locator.find(PreferencesAdapter.class);

    public int taxonomyLoadCount = 0;
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final Comparator<Taxon> taxonComparator = new Comparator<Taxon>()
      {
        public int compare (final @Nonnull Taxon taxon1, @Nonnull final Taxon taxon2)
          {
            return taxonCollator.compare(getDisplayName(taxon1), getDisplayName(taxon2));
          }

        @Nonnull
        private String getDisplayName (final @Nonnull Taxon taxon)
          {
            try
              {
                return taxon.getDisplayName(primaryTaxonLocale);
              }
            catch (NotFoundException e)
              {
                // We prepend some zeds to scientific names, so if they are compared against common names,
                // they go to the bottom of the list.
                return "ZZZZZZZ" + taxon.getScientificName();
              }
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public TaxonomyPreferencesSupport()
      {
        matcherMapByKey.put("FastNameMatcher", new FastNameMatcher(false));
        matcherMapByKey.put("FastNameMatcherWithInitial", new FastNameMatcher(true));
        matcherMapByKey.put("CamelCaseMatcher", new CamelCaseMatcher());
        matcherMapByKey.put("ExactSubstringNameMatcher", new ExactSubstringNameMatcher());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public NameMatcher getNameMatcher()
      {
        return nameMatcher;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Comparator<Taxon> getTaxonComparator()
      {
        return taxonComparator;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public String getSelectedTaxononyName()
      {
        return preferencesAdapter.getString(PREF_TAXONOMY, "EBNItalia2003");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public synchronized Taxonomy getTaxonomy()
      {
        if (taxonomy == null)
          {
            try
              {
                taxonomyLoadCount++;
                taxonomy = loadTaxonomy();
              }
            catch (Exception e)
              {
                throw new RuntimeException(e);
              }
          }

        return taxonomy;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void resetTaxonomy()
      {
        logger.info("resetTaxonomy()");
        taxonomy = null;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Taxonomy reloadTaxonomy()
      {
        logger.info("reloadTaxonomy()");
        taxonomy = null;
        final ObservationClipboard observationClipboard = Locator.find(ObservationClipboard.class);
        observationClipboard.clear();
        final TaxonHistory taxonHistory = Locator.find(TaxonHistory.class);
        taxonHistory.clearTaxonHistory();
        Locator.find(ObservationManager.class).findOrCreateObservationSetById(null).clear();
        
        return getTaxonomy();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public List<Locale> getTaxonomyLocales()
      {
        return Collections.unmodifiableList(taxonLocales);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isScientificNamesRenderingEnabled()
      {
        return preferencesAdapter.getBoolean(PREF_RENDER_SCIENTIFIC_NAMES, true);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void reset()
      {
        computeLocales();
        setNameMatcher();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected abstract Taxonomy loadTaxonomy()
      throws Exception;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected Taxonomy loadTaxonomy (final @Nonnull InputStream is, final @Nonnull String charSet)
      throws Exception
      {
        logger.info("loadTaxonomy(%s, %s)", is, charSet);
        Reader r = null;

        try
          {
            final long time = System.currentTimeMillis();
            final TaxonomyFactory taxonomyFactory = new SimpleTaxonomyFactory();
            r = new InputStreamReader(is, charSet);
            final Taxonomy taxonomy = taxonomyFactory.loadTaxonomy(r);
            logger.info(">>>> taxonomy loaded in %s msecs", System.currentTimeMillis() - time);
            return taxonomy;
          }
        finally
          {
            try
              {
                if (r != null)
                  {
                    r.close();
                  }
              }
            catch (IOException e)
              {
                logger.info("Loading the taxonomy: %s", e);
                logger.throwing("loadTaxonomy()", CLASS, e);
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void computeLocales()
      {
        logger.info("computeLocales() - default is %s", Locale.getDefault());
        primaryTaxonLocale = getLocale(PREF_PRIMARY_LOCALE, Locale.getDefault());
        final Locale secondaryTaxonLocale = getLocale(PREF_SECONDARY_LOCALE, primaryTaxonLocale);
        taxonLocales.clear();
        taxonLocales.add(primaryTaxonLocale);

        if (!primaryTaxonLocale.equals(secondaryTaxonLocale))
          {
            taxonLocales.add(secondaryTaxonLocale);
          }

        logger.info(">>>> locales: %s, primary: %s", taxonLocales, primaryTaxonLocale);
        taxonCollator = Collator.getInstance(primaryTaxonLocale);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected Locale getLocale (final @Nonnull String key, final @Nonnull Locale defaultLocale)
      {
        final String localeCode = preferencesAdapter.getString(key, "");
        return localeCode.equals("") ? defaultLocale : new Locale(localeCode);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void setNameMatcher()
      {
        final String key = preferencesAdapter.getString(PREF_NAME_MATCHER, DEFAULT_NAME_MATCHER);
        nameMatcher = matcherMapByKey.get(key);

        if (nameMatcher == null) // defensive
          {
            logger.warning("Cannot set name matcher %s - available: %s - using a default", key, matcherMapByKey.keySet());
            nameMatcher = new FastNameMatcher(true);
          }
      }
  }
